/*
 * arch/arm/mach-at91/pm.c
 * AT91 Power Management
 *
 * Copyright (C) 2005 David Brownell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/pm.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/pm.h>
#include <linux/interrupt.h>
#include <linux/sysfs.h>
#include <linux/module.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/atomic.h>
#include <asm/mach/time.h>
#include <asm/mach/irq.h>
#include <asm/mach-types.h>

#include <asm/arch/at91_pmc.h>
#include <asm/arch/gpio.h>
#include <asm/arch/cpu.h>

#include "generic.h"

/* SDRAMC self-refresh mode is entered a bit differently on RM9200 and SAM926x. */ 
#ifdef CONFIG_ARCH_AT91RM9200
/* 
 * On RM9200, we just trigger going into SR mode when all buffers flushed. 
 */

#include <asm/arch/at91rm9200_mc.h>

#define sdram_selfrefresh_save_enable()	at91_sys_write(AT91_SDRAMC_SRR, 1)
#define sdram_selfrefresh_restore() do {} while(0)

#else /* AT91SAM926x */
/* 
 * The SAM926x SDRAMC is capable of going into SR mode automatically when idle. 
 * We reconfigure the SDRAMC to minimal timeout and later restore previous setting.
 */

#include <asm/arch/at91sam926x_mc.h>

static u32 saved_lpr;

#define sdram_selfrefresh_save_enable()						\
	do {									\
		saved_lpr = at91_sys_read(AT91_SDRAMC_LPR);				\
		at91_sys_write(AT91_SDRAMC_LPR, AT91_SDRAMC_LPCB_SELF_REFRESH);	\
	} while (0);

#define sdram_selfrefresh_restore()    at91_sys_write(AT91_SDRAMC_LPR, saved_lpr)

#endif

#ifdef CONFIG_AT91_SLOW_CLOCK
/* Slow clock handler in pm_slowclock.S */
extern void at91_slow_clock(void);
/* Slow clock handler length */
extern u32 at91_slow_clock_sz;
#endif

static int at91_pm_valid_state(suspend_state_t state)
{
	switch (state) {
		case PM_SUSPEND_ON:
		case PM_SUSPEND_STANDBY:
		case PM_SUSPEND_MEM:
			return 1;

		default:
			return 0;
	}
}


static suspend_state_t target_state;

/*
 * Called after processes are frozen, but before we shutdown devices.
 */
static int at91_pm_set_target(suspend_state_t state)
{
	target_state = state;
	return 0;
}

/*
 * Verify that all the clocks are correct before entering
 * slow-clock mode.
 */
#warning "SAM9260 only has 3 programmable clocks."
static int at91_pm_verify_clocks(void)
{
	unsigned long scsr;
	int i;

	scsr = at91_sys_read(AT91_PMC_SCSR);

	/* USB must not be using PLLB */
	if (cpu_is_at91rm9200()) {
		if ((scsr & (AT91RM9200_PMC_UHP | AT91RM9200_PMC_UDP)) != 0) {
			pr_debug("AT91: PM - Suspend-to-RAM with USB still active\n");
			return 0;
		}
	} else if (cpu_is_at91sam9260() || cpu_is_at91sam9261() || cpu_is_at91sam9263()) {
		if ((scsr & (AT91SAM926x_PMC_UHP | AT91SAM926x_PMC_UDP)) != 0) {
			pr_debug("AT91: PM - Suspend-to-RAM with USB still active\n");
			return 0;
		}
	}

#ifdef CONFIG_AT91_PROGRAMMABLE_CLOCKS
	/* PCK0..PCK3 must be disabled, or configured to use clk32k */
	for (i = 0; i < 4; i++) {
		u32 css;

		if ((scsr & (AT91_PMC_PCK0 << i)) == 0)
			continue;

		css = at91_sys_read(AT91_PMC_PCKR(i)) & AT91_PMC_CSS;
		if (css != AT91_PMC_CSS_SLOW) {
			pr_debug("AT91: PM - Suspend-to-RAM with PCK%d src %d\n", i, css);
			return 0;
		}
	}
#endif

	return 1;
}

/*
 * This is called from clk_must_disable(), to see how deeply to suspend.
 * For example, some controllers (like OHCI) need one of the PLL clocks
 * in order to act as a wakeup source, and those are not available when
 * going into slow clock mode.
 */
int at91_suspend_entering_slow_clock(void)
{
	return (target_state == PM_SUSPEND_MEM);
}


#ifdef CONFIG_AT91_SLOW_CLOCK
static void (*slow_clock)(void);
#endif


static int at91_pm_enter(suspend_state_t state)
{
	at91_gpio_suspend();
	at91_irq_suspend();

	pr_debug("AT91: PM - wake mask %08x, pm state %d\n",
			/* remember all the always-wake irqs */
			(at91_sys_read(AT91_PMC_PCSR)
					| (1 << AT91_ID_FIQ)
					| (1 << AT91_ID_SYS)
					| (at91_extern_irq))
				& at91_sys_read(AT91_AIC_IMR),
			state);

	switch (state) {
		/*
		 * Suspend-to-RAM is like STANDBY plus slow clock mode, so
		 * drivers must suspend more deeply:  only the master clock
		 * controller may be using the main oscillator.
		 */
		case PM_SUSPEND_MEM:
			/*
			 * Ensure that clocks are in a valid state.
			 */
			if (!at91_pm_verify_clocks())
				goto error;

#ifdef CONFIG_AT91_SLOW_CLOCK
			/* 
			 * First, copy sclk handler to SRAM. This may overwrite 
			 * some of framebuffer etc.
			 * REVISIT: the SRAM memory area under slow clock handler
			 * should be backed up to SDRAM and restored after suspend!  
			 */
			memcpy(slow_clock, at91_slow_clock, at91_slow_clock_sz);
			slow_clock();
			break;
#else
			pr_info("AT91: PM - no slow clock mode ...\n");
			/* FALLTHROUGH leaving master clock alone */
#endif

		/*
		 * STANDBY mode has *all* drivers suspended; ignores irqs not
		 * marked as 'wakeup' event sources; and reduces DRAM power.
		 * But otherwise it's identical to PM_SUSPEND_ON:  cpu idle, and
		 * nothing fancy done with main or cpu clocks.
		 */
		case PM_SUSPEND_STANDBY:
			/*
			 * NOTE: the Wait-for-Interrupt instruction needs to be
			 * in icache so the SDRAM stays in self-refresh mode until
			 * the wakeup IRQ occurs.
			 * The RM9200 goes into SR mode for a command and stays until
			 * first access. The SAM926x goes into SR mode always after some 
			 * delay after the last access and stays until the next. 
			 * SAM926x may have the automatic SR enabled always to save power.
			 * Here, we change the delay to minimum and later restore it
			 * as it was before.    
			 */
			asm("b 1f; .align 5; 1:");
			asm("mcr p15, 0, r0, c7, c10, 4");	/* drain write buffer */

			/* Enable the self-refresh mode for SDRAM to lower the power consumption */
			sdram_selfrefresh_save_enable();
			
			asm("mcr p15, 0, r0, c7, c0, 4");	/* wait for interrupt */

			/* On AT91SAM, configure the low-power mode back to as it was before*/
			sdram_selfrefresh_restore();
			break;
			

		case PM_SUSPEND_ON:
			asm("mcr p15, 0, r0, c7, c0, 4");	/* wait for interrupt */
			break;

		default:
			pr_debug("AT91: PM - bogus suspend state %d\n", state);
			goto error;
	}

	pr_debug("AT91: PM - wakeup %08x\n",
			at91_sys_read(AT91_AIC_IPR) & at91_sys_read(AT91_AIC_IMR));

error:
	target_state = PM_SUSPEND_ON;
	at91_irq_resume();
	at91_gpio_resume();
	return 0;
}


static struct pm_ops at91_pm_ops ={
	.valid		= at91_pm_valid_state,
	.set_target	= at91_pm_set_target,
	.enter		= at91_pm_enter,
};

static int __init at91_pm_init(void)
{
#ifdef CONFIG_AT91_SLOW_CLOCK
	/*
	 * REVISIT allocations of SRAM should be dynamically managed.
	 * FIQ handlers and other components will want SRAM/TCM too...
	 * Or just allocate SDRAM and back it up before overwriting
	 * and restore after return from standby (so that whatever was
	 * there, nobody would notice it has been gone awhile).
	 */

	/* 
	 * Handler is copied to SRAM top, above framebuffer. This way, we don't
	 * mess up the framebuffer when going to slow clock mode.
	 * Actual copying takes place later, just before going to slow clock mode.
	 * Otherwise, if something has overwritten the handler, we're in trouble.
	 */
	slow_clock = (void *) (AT91_IO_VIRT_BASE - at91_slow_clock_sz);

	printk("AT91: Power Management (with slow clock mode)\n");
#else
	printk("AT91: Power Management\n");
#endif

#ifdef CONFIG_ARCH_AT91RM9200
	/* Disable RM9200 SDRAM low-power mode. Cannot be used with self-refresh. */
	at91_sys_write(AT91_SDRAMC_LPR, 0);
#endif

	pm_set_ops(&at91_pm_ops);

	return 0;
}
arch_initcall(at91_pm_init);
