/*
 * arch/arm/mach-at91/pm_slow_clock.S
 *
 *  Copyright (C) 2006 Savin Zlobec
 * AT91SAM support by:
 *  Copyright (C) 2007 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This has been tested on AT91SAM9261 with over 1,8M suspend cycles
 */

#include <linux/linkage.h>
#include <asm/hardware.h>
#include <asm/arch/at91_pmc.h>

#ifdef CONFIG_ARCH_AT91RM9200
#include <asm/arch/at91rm9200_mc.h>
#else
#include <asm/arch/at91sam926x_mc.h>
#endif

/* Allow timeouts when waiting for lock and ready bits */
/* Does not change much, because we're screwed anyway if something fails */
//#define USE_TIMEOUTS

#ifdef USE_TIMEOUTS

#define MCKRDY_TIMEOUT		1000
#define MOSCRDY_TIMEOUT 	1000
#define PLLALOCK_TIMEOUT	1000
#define PLLBLOCK_TIMEOUT	1000

#endif

/* ---------------------------------------------------------------------------- */
/* MACROS */
/* ---------------------------------------------------------------------------- */
/* Wait until master clock is ready (after switching master clock sources) */
	.macro wait_mckrdy
#ifdef USE_TIMEOUTS
	mov	r2, #MCKRDY_TIMEOUT
1:	sub	r2, r2, #1
	cmp	r2, #0
	beq	2f
#else
1:
#endif
	ldr	r3, [r1, #(AT91_PMC_SR - AT91_PMC)]
	tst	r3, #AT91_PMC_MCKRDY
	beq	1b
2:
	.endm

/* ---------------------------------------------------------------------------- */
/* Wait until master oscillator has stabilized */
	.macro wait_moscrdy
#ifdef USE_TIMEOUTS
	mov	r2, #MOSCRDY_TIMEOUT
1:	sub	r2, r2, #1
	cmp	r2, #0
	beq	2f
#else
1:
#endif
	ldr	r3, [r1, #(AT91_PMC_SR - AT91_PMC)]
	tst	r3, #AT91_PMC_MOSCS
	beq	1b
2:
	.endm

/* ---------------------------------------------------------------------------- */
/* Wait until PLLA has locked*/
	.macro wait_pllalock
#ifdef USE_TIMEOUTS	
	mov	r2, #PLLALOCK_TIMEOUT
1:	sub	r2, r2, #1
	cmp	r2, #0
	beq	2f
#else
1:
#endif
	ldr	r3, [r1, #(AT91_PMC_SR - AT91_PMC)]
	tst	r3, #AT91_PMC_LOCKA
	beq	1b
2:
	.endm

/* ---------------------------------------------------------------------------- */
/* Wait until PLLB has locked*/
	.macro wait_pllblock
#ifdef USE_TIMEOUTS	
	mov	r2, #PLLBLOCK_TIMEOUT
1:	sub	r2, r2, #1
	cmp	r2, #0
	beq	2f
#else
1:
#endif
	ldr	r3, [r1, #(AT91_PMC_SR - AT91_PMC)]
	tst	r3, #AT91_PMC_LOCKB
	beq	1b
2:
	.endm



/* ---------------------------------------------------------------------------- */
	.text	
/* ---------------------------------------------------------------------------- */
/* Called from pm.c to do slow-clock standby */
ENTRY(at91_slow_clock)

	/* save registers to stack */
	stmfd	sp!, {r0 - r12, lr}		
	

/* REGISTER USAGE:
 * r0 - constant zero
 * r1 - BASE_PMC
 * r2-r3 - inline usage / macros!!!
 * r4 - BASE_SDRAMC
 * r8-r11 - backed up register values 
 */


	/* ---------------------------------------------------------------------------- */
	/* Load SDRAMC VA base address to r2 */
	ldr	r4, .at91_va_base_sdramc

	/* Load PMC VA base address to r1 */
	ldr	r1, .at91_va_base_pmc

	/* Save running settings */
	ldr	r8, [r1, #(AT91_PMC_MCKR - AT91_PMC)]

	ldr	r9, [r1, #(AT91_CKGR_PLLAR - AT91_PMC)]

	ldr	r10, [r1, #(AT91_CKGR_PLLBR - AT91_PMC)]
	
	ldr	r12, [r1, #(AT91_CKGR_MOR - AT91_PMC)]
	
	/* Put SDRAM in self refresh mode - AT91SAM */
#ifndef CONFIG_ARCH_AT91RM9200
	/* Save LPR contents, so we could restore it later */
	ldr	r11, [r4, #(AT91_SDRAMC_LPR - AT91_SDRAMC)]
	
	/* Enable SDRAM Self-Refresh when SDRAMC idle */
	mov	r2, #AT91_SDRAMC_LPCB_SELF_REFRESH
	str	r2, [r4, #(AT91_SDRAMC_LPR - AT91_SDRAMC)]
	
	/* On AT91SAM, the sdram goes to self-refresh automatically after all
	operations have completed. Enable the self-refresh mode before draining
	buffers. */
#endif

	b	1f
	.align	5
	
	/* Drain write buffer */
	mov	r0, #0
1:	mcr	p15, 0, r0, c7, c10, 4
	nop

	/* Put SDRAM in self refresh mode - AT91RM9200 */
#ifdef CONFIG_ARCH_AT91RM9200
	mov	r2, #1
	str	r2, [r4, #(AT91_SDRAMC_SRR - AT91_MC)]
	/* On RM9200, the sdram goes to self-refresh mode immediately on this command */
#endif

	/* ---------------------------------------------------------------------------- */

	/* STEP 1: Switch clock mux from fast clock (usually PLLA) to slow clock */ 
	bic	r2, r8, #3
	str	r2, [r1, #(AT91_PMC_MCKR - AT91_PMC)]
	wait_mckrdy

	/* Originally, we changed clock divider back to 1, too. (in a separate step here) 
	 * This caused some strange behaviour (like instability - some bits of a word were
	 * not written to PMC registers for a short period after changing divider etc) 
	 * about once in 3000 suspend cycles on AT91SAM9261 rev A. A extra  delay of 1ms
	 * helped a bit - then it crashed only once in 100k cycles.
	 * The idea was to switch the cpu down to 500Hz but with given problems, it
	 * does not seem wise to touch MCKR more than minimally needed (CSS -> slow clock). 
	 */

	/* STEP 2a: Disable PLLA */
	mov	r2, #0x3F00
	orr r2, r2, #(1<<29)
	str	r2, [r1, #(AT91_CKGR_PLLAR - AT91_PMC)]

	/* STEP 2b: Disable PLLB */
	mov	r2, #0x3F00
	str	r2, [r1, #(AT91_CKGR_PLLBR - AT91_PMC)]

	/* STEP 3: Turn off the main oscillator */
	bic	r2, r12, #AT91_PMC_MOSCEN
	str	r2, [r1, #(AT91_CKGR_MOR - AT91_PMC)]

	/* ---------------------------------------------------------------------------- */
	/* EVERYTHING SHOULD BE READY FOR SLOW CLOCK MODE HERE */
	/* Wait for interrupt to wake us up */

	mcr	p15, 0, r0, c7, c0, 4

	/* We're back */
	/* RESTORE CLOCK SETTINGS */

	/* STEP 1: Turn on main osc and wait for stabilization*/
	str	r12, [r1, #(AT91_CKGR_MOR - AT91_PMC)]
	wait_moscrdy

	/* STEP 2a: Restore PLLA setting and wait for PLLA lock signal */
	/*  (only if PLLA was used) */
	tst r9, #0xFF
	beq 3f

	str	r9, [r1, #(AT91_CKGR_PLLAR - AT91_PMC)]
	wait_pllalock
3:

	/* STEP 2b: Restore PLLB setting and wait for PLLB lock signal  */
	/*  (only if PLLB was used) */
	tst r10, #0xFF
	beq 3f
	
	str	r10, [r1, #(AT91_CKGR_PLLBR - AT91_PMC)]
	wait_pllblock
3:

	/* STEP 3: switch clock mux back to fast clock (usually PLLA) */
	/* We've changed only CSS bits of MCKR, so doing this with single write is OK */
	str	r8, [r1, #(AT91_PMC_MCKR - AT91_PMC)]
	wait_mckrdy
	

#ifndef CONFIG_ARCH_AT91RM9200
	/* Restore LPR on AT91SAM */
	str	r11, [r4, #(AT91_SDRAMC_LPR - AT91_SDRAMC)]
#endif

	ldmfd	sp!, {r0 - r12, pc}		@ restore regs and return

	/* ---------------------------------------------------------------------------- */
	/* Variables */

	/* The virtual base addresses are given to PMC and SDRAMC separately.
	 * ARM allows with base+offset addressing the offset to be only 8-bit. 
	 * When refrerenced to BASE_SYS, the offsets would grow too large. 
	 */
.at91_va_base_pmc:
	.word (AT91_VA_BASE_SYS + AT91_PMC)

.at91_va_base_sdramc:
#ifdef CONFIG_ARCH_AT91RM9200
	.word (AT91_VA_BASE_SYS + AT91_MC)
#else
	.word (AT91_VA_BASE_SYS + AT91_SDRAMC)
#endif

	/* ---------------------------------------------------------------------------- */
	/* Length of the at91_slow_clock function. */
	/* Needed to copy it to SRAM before use (see pm.c). */
ENTRY(at91_slow_clock_sz)
	.word .-at91_slow_clock
