/*
 *
 *	Panelcard Base matrix keyboard driver
 *
 *	Copyright (C) 2007 taskit GmbH
 *
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>

#include <linux/delay.h>

#include <linux/platform_device.h>

#include <asm/arch/gpio.h>

#include <linux/at91-matrix.h>

#define AT91_MATRIX_NAME	"AT91 matrix driver"

static void timer_handler(unsigned long data);

struct at91_matrix_t {
	int scanning;
	char key_flags[MAX_COLS][MAX_ROWS];
	struct timer_list timer;
	struct input_dev *input;
	struct at91_matrix_data *data;
};

static struct at91_matrix_t at91_matrix = {
	.scanning = 0,
	.timer    = TIMER_INITIALIZER(&timer_handler, 0, 0),
	.input    = NULL,
	.data     = NULL
};

/*
 *	scan routine
 *	reports first key found
 */
static int scan_matrix(void)
{
	int x, y;
	int key_found = 0;

	/* rows are already configurated as inputs with pull-up */

	/* set high level for all cols */
	for(x = 0; (x < MAX_COLS) && at91_matrix.data->col[x]; x++)
		at91_set_gpio_value(at91_matrix.data->col[x], 1);

	for(x = 0; (x < MAX_COLS) && at91_matrix.data->col[x]; x++) {
		/* one col is low */
		at91_set_gpio_value(at91_matrix.data->col[x], 0);
		udelay(5);

		/* scan rows */
		for(y = 0; (y < MAX_ROWS) && at91_matrix.data->row[y]; y++) {
			if(at91_get_gpio_value(at91_matrix.data->row[y]) == 0) {
				key_found = 1;
				if(at91_matrix.key_flags[y][x] == 0) {
					/* key was unpressed -> report key press */
					at91_matrix.key_flags[y][x] = 1;
					input_report_key(at91_matrix.input, at91_matrix.data->keymap[y][x], 1);
					input_sync(at91_matrix.input);
				}
			} else {
				if(at91_matrix.key_flags[y][x]) {
					/* key was pressed and is now unpressed -> report key release */
					at91_matrix.key_flags[y][x] = 0;
					input_report_key(at91_matrix.input, at91_matrix.data->keymap[y][x], 0);
					input_sync(at91_matrix.input);
				}
			}
		}

		/* restore high level */
		at91_set_gpio_value(at91_matrix.data->col[x], 1);
	}

	return(key_found);
}

static void timer_handler(unsigned long data)
{
	int i;

	/* reschedule scan if any key is pressed */
	if(scan_matrix()) {
		/* rescan matrix in 25 ms */
		at91_matrix.timer.expires = jiffies + 25 * HZ / 1000;
		add_timer(&at91_matrix.timer);
	} else {
		/* output low on all cols */
		for(i = 0; (i < MAX_COLS) && at91_matrix.data->col[i]; i++)
			at91_set_gpio_value(at91_matrix.data->col[i], 0);
		/* reenable interrupt */
		at91_matrix.scanning = 0;
	}
}

/*
 *	interrupt handler
 *	if any key is pressed scan_matrix() is invoked in non-interrupt context
 */
static irqreturn_t irqhandler(int irq, void *dev_id)
{
	if(at91_matrix.data && !at91_matrix.scanning) {
		at91_matrix.scanning = 1;
		/* start scanning in 10 ms */
		at91_matrix.timer.expires = jiffies + 10 * HZ / 1000;
		add_timer(&at91_matrix.timer);
	}
	return(IRQ_HANDLED);
}


static void at91_matrix_free_irqs(struct at91_matrix_data *data, int n)
{
	int i;

	for(i = 0; (i < n) && data->row[i]; i++)
		free_irq(data->row[i], &at91_matrix);
}

static int at91_matrix_probe(struct platform_device *pdev)
{
	struct at91_matrix_data *data;
	int i;

	data = pdev->dev.platform_data;
	if(data == NULL)
		return(-1);

	memset(at91_matrix.key_flags, 0, sizeof(at91_matrix.key_flags));

	/* configure all rows as input with pull-up */
	for(i = 0; (i < MAX_ROWS) && data->row[i]; i++)
		at91_set_gpio_input(data->row[i], 1);

	/* configure all cols as low level multidrive with pull-up */
	for(i = 0; (i < MAX_COLS) && data->col[i]; i++) {
		/* pull-up on rows provide high level */
		at91_set_multi_drive(data->col[i], 1);
		at91_set_gpio_output(data->col[i], 0);
	}

	/* request IRQ for each line */
	for(i = 0; (i < MAX_ROWS) && data->row[i]; i++)
		if(request_irq(data->row[i], &irqhandler, 0, AT91_MATRIX_NAME, &at91_matrix)) {
			at91_matrix_free_irqs(data, i);
			return(-1);
		}

	/* register input device */
	at91_matrix.input = input_allocate_device();
	if(at91_matrix.input == NULL) {
		at91_matrix_free_irqs(data, MAX_ROWS);
		return(-1);
	}
	at91_matrix.input->name = "AT91 matrix keyboard";
	at91_matrix.input->evbit[0] = BIT(EV_KEY);
	for(i = 0; i < KEY_MAX; i++)
		set_bit(i, at91_matrix.input->keybit);
	input_register_device(at91_matrix.input);

	at91_matrix.data = data;

	return(0);
}

static int at91_matrix_remove(struct platform_device *pdev)
{
	struct at91_matrix_data *data;

	data = pdev->dev.platform_data;
	if(data == NULL)
		return(-1);

	at91_matrix_free_irqs(data, MAX_ROWS);

	del_timer_sync(&at91_matrix.timer);

	if(at91_matrix.input) {
		input_unregister_device(at91_matrix.input);
	}

	return(0);
}

static struct platform_driver at91_matrix_driver = {
	.probe    = at91_matrix_probe,
	.remove   = at91_matrix_remove,
	.suspend  = NULL,
	.resume   = NULL,
	.driver   = {
		.name = "at91_matrix",
	},
};

static int __init at91_matrix_init(void)
{
	return platform_driver_register(&at91_matrix_driver);
}

static void __exit at91_matrix_exit(void)
{
 	platform_driver_unregister(&at91_matrix_driver);
}

module_init(at91_matrix_init)
module_exit(at91_matrix_exit)

MODULE_AUTHOR("Mathias Langer");
MODULE_DESCRIPTION("Panelcard matrix keyboard driver");
MODULE_LICENSE("GPL v2");
