/*
 * Portux Panel-PC PS/2 keyboard submodule
 *
 * Copyright (C) 2005-2007 taskit GmbH
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/portuxinput.h>

static unsigned char atkbd_set2_keycode[512] = {
	  0, 67, 65, 63, 61, 59, 60, 88,  0, 68, 66, 64, 62, 15, 41,117,
	  0, 56, 42, 93, 29, 16,  2,  0,  0,  0, 44, 31, 30, 17,  3,  0,
	  0, 46, 45, 32, 18,  5,  4, 95,  0, 57, 47, 33, 20, 19,  6,183,
	  0, 49, 48, 35, 34, 21,  7,184,  0,  0, 50, 36, 22,  8,  9,185,
	  0, 51, 37, 23, 24, 11, 10,  0,  0, 52, 53, 38, 39, 25, 12,  0,
	  0, 89, 40,  0, 26, 13,  0,  0, 58, 54, 28, 27,  0, 43,  0, 85,
	  0, 86, 91, 90, 92,  0, 14, 94,  0, 79,124, 75, 71,121,  0,  0,
	 82, 83, 80, 76, 77, 72,  1, 69, 87, 78, 81, 74, 55, 73, 70, 99,

	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	217,100,255,  0, 97,165,  0,  0,156,  0,  0,  0,  0,  0,  0,125,
	173,114,  0,113,  0,  0,  0,126,128,  0,  0,140,  0,  0,  0,127,
	159,  0,115,  0,164,  0,  0,116,158,  0,150,166,  0,  0,  0,142,
	157,  0,  0,  0,  0,  0,  0,  0,155,  0, 98,  0,  0,163,  0,  0,
	226,  0,  0,  0,  0,  0,  0,  0,  0,255, 96,  0,  0,  0,143,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,107,  0,105,102,  0,  0,112,
	110,111,108,112,106,103,  0,119,  0,118,109,  0, 99,104,119,  0,

	  0,  0,  0, 65, 99,
};

#define ATKBD_RET_BAT		0xaa
#define ATKBD_RET_EMUL0		0xe0
#define ATKBD_RET_EMUL1		0xe1
#define ATKBD_RET_RELEASE	0xf0
#define ATKBD_RET_HANGUEL	0xf1
#define ATKBD_RET_HANJA		0xf2
#define ATKBD_RET_ERR		0xff

#define ATKBD_KEY_UNKNOWN	  0
#define ATKBD_KEY_NULL		255

#define ATKBD_SCR_1		254
#define ATKBD_SCR_2		253
#define ATKBD_SCR_4		252
#define ATKBD_SCR_8		251
#define ATKBD_SCR_CLICK		250
#define ATKBD_SCR_LEFT		249
#define ATKBD_SCR_RIGHT		248

/*
 * The atkbd control structure
 */

struct atkbd {

	struct input_dev *dev;

	unsigned short id;
	unsigned char keycode[512];
	unsigned char set;
	unsigned char softrepeat;
	unsigned char softraw;
	unsigned char scroll;

	/* Accessed only from interrupt */
	unsigned char emul;
	unsigned char resend;
	unsigned char release;
	unsigned char bat_xl;
	unsigned int last;
	unsigned long time;
} portuxkbd;

static void atkbd_report_key(struct input_dev *dev, int code, int value)
{
	if (value == 3) {
		input_report_key(dev, code, 1);
		input_sync(dev);
		input_report_key(dev, code, 0);
	} else
		input_event(dev, EV_KEY, code, value);
	input_sync(dev);
}

static void kbdhandler(char *data)
{
	struct atkbd *atkbd = &portuxkbd;
	unsigned int code = *data;
	int scroll = 0, hscroll = 0, click = -1;
	int value;

	switch (code) {
		case ATKBD_RET_BAT:
			goto out;
		case ATKBD_RET_EMUL0:
			atkbd->emul = 1;
			goto out;
		case ATKBD_RET_EMUL1:
			atkbd->emul = 2;
			goto out;
		case ATKBD_RET_RELEASE:
			atkbd->release = 1;
			goto out;
		case ATKBD_RET_HANGUEL:
			atkbd_report_key(atkbd->dev, KEY_HANGUEL, 3);
			goto out;
		case ATKBD_RET_HANJA:
			atkbd_report_key(atkbd->dev, KEY_HANJA, 3);
			goto out;
		case ATKBD_RET_ERR:
			goto out;
	}

	if (atkbd->set != 3)
		code = (code & 0x7f) | ((code & 0x80) << 1);
	if (atkbd->emul) {
		if (--atkbd->emul)
			goto out;
		code |= (atkbd->set != 3) ? 0x80 : 0x100;
	}

	switch (atkbd->keycode[code]) {
		case ATKBD_KEY_NULL:
			break;
		case ATKBD_KEY_UNKNOWN:
			input_sync(atkbd->dev);
			break;
		case ATKBD_SCR_1:
			scroll = 1 - atkbd->release * 2;
			break;
		case ATKBD_SCR_2:
			scroll = 2 - atkbd->release * 4;
			break;
		case ATKBD_SCR_4:
			scroll = 4 - atkbd->release * 8;
			break;
		case ATKBD_SCR_8:
			scroll = 8 - atkbd->release * 16;
			break;
		case ATKBD_SCR_CLICK:
			click = !atkbd->release;
			break;
		case ATKBD_SCR_LEFT:
			hscroll = -1;
			break;
		case ATKBD_SCR_RIGHT:
			hscroll = 1;
			break;
		default:
			value = atkbd->release ? 0 :
				(1 + (!atkbd->softrepeat && test_bit(atkbd->keycode[code], atkbd->dev->key)));
			atkbd_report_key(atkbd->dev, atkbd->keycode[code], value);
	}

	if (click != -1)
		input_report_key(atkbd->dev, BTN_MIDDLE, click);
	input_report_rel(atkbd->dev, REL_WHEEL, scroll);
	input_report_rel(atkbd->dev, REL_HWHEEL, hscroll);
	input_sync(atkbd->dev);

	atkbd->release = 0;
out:
	return;
}

static int portuxkbd_event(struct input_dev *dev, unsigned int type, unsigned int code, int value)
{
	struct atkbd *atkbd = dev->private;
	const short period[32] =
		{ 33,  37,  42,  46,  50,  54,  58,  63,  67,  75,  83,  92, 100, 109, 116, 125,
		 133, 149, 167, 182, 200, 217, 232, 250, 270, 303, 333, 370, 400, 435, 470, 500 };
	const short delay[4] =
		{ 250, 500, 750, 1000 };
	unsigned char param;
	int i, j;

	switch (type) {
	case EV_LED:
		param = (test_bit(LED_SCROLLL, dev->led) ? 1 : 0)
		      | (test_bit(LED_NUML,    dev->led) ? 2 : 0)
		      | (test_bit(LED_CAPSL,   dev->led) ? 4 : 0);
	        portuxinput_schedule_ps2command(0xed, param);
		return 0;

	case EV_REP:
		if (atkbd->softrepeat) return 0;

		i = j = 0;
		while (i < 31 && period[i] < dev->rep[REP_PERIOD])
			i++;
		while (j < 3 && delay[j] < dev->rep[REP_DELAY])
			j++;
		dev->rep[REP_PERIOD] = period[i];
		dev->rep[REP_DELAY] = delay[j];
		param = i | (j << 5);
	        portuxinput_schedule_ps2command(0xf3, param);
		return 0;
	}

	return -1;
}

static int __init portuxkbd_init(void)
{
	if(portuxinput_get_platform_data()) {
		int i;
		memset(&portuxkbd, 0, sizeof(struct atkbd));
		portuxkbd.dev=input_allocate_device();
		portuxkbd.dev->name="Portux ps2 keyboard";
		portuxkbd.dev->evbit[0]=BIT(EV_KEY)|BIT(EV_REL)|BIT(EV_LED);
		portuxkbd.dev->relbit[0]=BIT(REL_WHEEL)|BIT(REL_HWHEEL);
		portuxkbd.dev->ledbit[0] = BIT(LED_NUML) | BIT(LED_CAPSL) | BIT(LED_SCROLLL);
		portuxkbd.dev->event = portuxkbd_event;
		portuxkbd.set=2;
		portuxkbd.dev->private = &portuxkbd;
		memcpy(portuxkbd.keycode, atkbd_set2_keycode, sizeof(portuxkbd.keycode));
		for(i=0;i<KEY_MAX;i++)
			set_bit(i, portuxkbd.dev->keybit);
		portuxinput_register_kbd(&kbdhandler, NULL);
		input_register_device(portuxkbd.dev);
	}
	return 0;
}

static void __exit portuxkbd_exit(void)
{
	portuxinput_schedule_ps2command(0xed, 0);
	input_unregister_device(portuxkbd.dev);
	portuxinput_deregister_kbd(&kbdhandler);
}

module_init(portuxkbd_init)
module_exit(portuxkbd_exit)

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("PS/2 keyboard submodule for portuxinput");
MODULE_LICENSE("GPL v2");

