/*
 * Backlight driver for taskit's Panel-Card
 *
 * Copyright (C) 2007 taskit GmbH
 */
 
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/backlight.h>
#include <asm/arch-at91/gpio.h>

#include <linux/platform_device.h>

static int current_intensity = 0;
static int panelcardbl_suspended = 0;

#define LOW 0
#define HIGH 1
#define TRISTATE 2

static int brightness[][2]={{HIGH    , HIGH},
                            {TRISTATE, HIGH},
			    {LOW     , HIGH},
			    {HIGH    , TRISTATE},
			    {TRISTATE, TRISTATE},
			    {LOW     , TRISTATE}};

static void panelcardbl_set_pin(unsigned pin, int value) {
	switch(value) {
	case HIGH:
		at91_set_gpio_value(pin, 1);
		at91_set_multi_drive(pin, 0);
		break;
	case LOW:
		at91_set_gpio_value(pin, 0);
		break;
	case TRISTATE:
		at91_set_gpio_value(pin, 1);
		at91_set_multi_drive(pin, 1);
		break;
	}
}

static void panelcardbl_set_intensity(struct backlight_device *dev) {
	if(panelcardbl_suspended||dev->props.power != 0||dev->props.fb_blank != 0) {
		at91_set_gpio_value(AT91_PIN_PC1, 0);
		at91_set_gpio_value(AT91_PIN_PC4, 1);
		at91_set_gpio_value(AT91_PIN_PC5, 1);
	} else {
		at91_set_gpio_value(AT91_PIN_PC1, 1);
		panelcardbl_set_pin(AT91_PIN_PC4, brightness[dev->props.brightness][0]);
		panelcardbl_set_pin(AT91_PIN_PC5, brightness[dev->props.brightness][1]);
		
		current_intensity=dev->props.brightness;
	}	
}

static int panelcardbl_update_status(struct backlight_device *dev) {
	panelcardbl_set_intensity(dev);
	return 0;
}

static int panelcardbl_get_brightness(struct backlight_device *dev) {
	return current_intensity;
}

static struct backlight_device *panelcard_backlight_device;
static struct backlight_ops panelcard_backlight_properties={
	.update_status	= panelcardbl_update_status,
	.get_brightness	= panelcardbl_get_brightness,
};

#ifdef CONFIG_PM
static int panelcardbl_suspend(struct platform_device *dev, pm_message_t state)
{
	panelcardbl_suspended = 1;
	panelcardbl_set_intensity(panelcard_backlight_device);
	return 0;
}

static int panelcardbl_resume(struct platform_device *dev)
{
	panelcardbl_suspended = 0;
	panelcardbl_set_intensity(panelcard_backlight_device);
	return 0;
}
#else
#define panelcardbl_suspend	NULL
#define panelcardbl_resume	NULL
#endif


static int panelcardbl_probe(struct platform_device *dev)
{
	panelcard_backlight_device = backlight_device_register("panelcard-bl",
		&dev->dev, NULL, &panelcard_backlight_properties);
	if (IS_ERR(panelcard_backlight_device))
		return PTR_ERR (panelcard_backlight_device);

	at91_set_gpio_output(AT91_PIN_PC1, 0);
	at91_set_gpio_output(AT91_PIN_PC4, 1);
	at91_set_gpio_output(AT91_PIN_PC5, 1);

	panelcard_backlight_device->props.brightness = current_intensity = 3;
	panelcard_backlight_device->props.max_brightness = sizeof(brightness)/sizeof(brightness[0])-1;
	panelcardbl_set_intensity(panelcard_backlight_device);

	return 0;
}

static int panelcardbl_remove(struct platform_device *dev)
{
	backlight_device_unregister(panelcard_backlight_device);

	return 0;
}

static struct platform_device *panelcardbl_device;
static struct platform_driver panelcardbl_driver = {
	.probe		= panelcardbl_probe,
	.remove		= panelcardbl_remove,
	.suspend	= panelcardbl_suspend,
	.resume		= panelcardbl_resume,
	.driver		= {
		.name	= "panelcard-bl",
	},
};


static int __init panelcardbl_init(void)
{
	int ret;

	ret = platform_driver_register(&panelcardbl_driver);
	if (!ret) {
		panelcardbl_device = platform_device_alloc("panelcard-bl", -1);
		if (!panelcardbl_device)
			return -ENOMEM;

		ret = platform_device_add(panelcardbl_device);

		if (ret) {
			platform_device_put(panelcardbl_device);
			platform_driver_unregister(&panelcardbl_driver);
		}
	}
	return ret;
}

static void __exit panelcardbl_exit(void)
{
	platform_device_unregister(panelcardbl_device);
 	platform_driver_unregister(&panelcardbl_driver);
}

module_init(panelcardbl_init);
module_exit(panelcardbl_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Christian Glindkamp");
