//-----------------------------------------------------------------------------
//  
// linux/drivers/video/epson/s1d13xxxfb_clock.c -- Clock Chip support  
// functions for Epson S1D13xxx controller frame buffer drivers.
// 
// Copyright(c) Seiko Epson Corporation 2000-2005.
// All rights reserved.
//
// This file is subject to the terms and conditions of the GNU General Public
// License. See the file COPYING in the main directory of this archive for
// more details.
//
//---------------------------------------------------------------------------- 
//---------------------------------------------------------------------------- 
//
// WARNING: This file uses a chip specific include file. The include file 
// contains all hardware information needed to generate proper timings, color
// depth, screen resolution etc.
//
// This driver supports one of the following Epson CRT/TV/LCD controllers:
//
// S1D13A05
//
//-----------------------------------------------------------------------------

#include <linux/kernel.h>
#include "s1d13xxxfb.h"

#ifdef CONFIG_FBCON_EPSON_CLOCK_CHIP

// Clock frequencies in Hz to be supplied in programming the clock chip.
// These values are obtained by using the XXXCFG.EXE tool.  These
// specific frequencies are for the Sharp LG038Q5DR01 panel.
#define S1D13XXX_CLKI       50000000
#define S1D13XXX_CLKI2      6800000

// GPIO Status and Control Register
#define REG0064_GPIO_STATUS_CONTROL         0x64

//-----------------------------------------------------------------------------
//
// Local Definitions
// 
//---------------------------------------------------------------------------

// Clock Chip Support
typedef enum
{
	ICD_VREG0 = 0,
	ICD_VREG1,
	ICD_VREG2,
	ICD_MREG
} ICD_REG;

typedef struct tagICD_CLK_INFO
{
	u32 ulReqFreq;                   // Requested frequency.
	int M;                              // Post divisor value (0 < D < 15)
	int I;                              // VCO index value. (see dwIndexFreq)
	int P, Q;                           // P and Q counter values
	int fVCO;                           //
	u32 ulBits;                      // Holder to build the bit pattern into.
} ICD_CLK_INFO, *pICD_CLK_INFO;

//-----------------------------------------------------------------------------
//
// Function Prototypes 
// 
//-----------------------------------------------------------------------------
int s1d13xxxfb_SetClock(void);
static int s1d13xxxfb_pCalcClockBits( ICD_REG Reg, u32 ulDesiredFreq, u32 * pulBits );
static int s1d13xxxfb_pProgramClockChip( u32 ulBits );
static void WriteClockEdge( int FallingEdgeBit, int RisingEdgeBit );

//-----------------------------------------------------------------------------
//
// Local globals
//
//----------------------------------------------------------------------------- 

#ifdef CONFIG_FBCON_EPSON_S1D13A05
// Clock Chip Support
static u32 ulIndexFreq[16][2] =
{
	{50000000,  51000000},
	{51000000,  53200000},
	{53200000,  58500000},
	{58500000,  60700000}, 
	{60700000,  64400000},
	{64400000,  66800000},
	{66800000,  73500000},
	{73500000,  75600000}, 
	{75600000,  80900000},
	{80900000,  83200000},
	{83200000,  91500000},
	{91500000,  100000000}, 
	{100000000, 120000000},
	{100000000, 120000000},
	{100000000, 120000000},     // Turn off VCLK
	{100000000, 120000000},     // Mux MCLK to VCLK
};

static u32 fREF = 14318180;  // 14.31818 MHz
#endif

/*=========================================================================*/
/*                           Programable Clock                             */
/*=========================================================================*/

//---------------------------------------------------------------------------
//  FUNCTION:   s1d13xxxfb_SetClock()
//
//  DESCRIPTION:
//      This routine will program the ICD2061A programmable clock generator
//      to the specifed frequencies.
//
//      NOTE:   The actual frequencies may not exactly match the requested
//              frequency.
//
//  PARAMETERS:
//
//  RETURNS:
//      1 if the function was successful in setting the clock.
//      0 if there was an error detected while trying to set the clock.
//
//  MODIFIES:
//---------------------------------------------------------------------------
int s1d13xxxfb_SetClock( void )
{
	u32 ulBits;

	// RELE - do a little sanity check on the state variables.

	if (0 != S1D13XXX_CLKI) {
		if (0 == s1d13xxxfb_pCalcClockBits(ICD_VREG0, S1D13XXX_CLKI, &ulBits))
			return 0;
		if (0 == s1d13xxxfb_pProgramClockChip(ulBits))
			return 0;

		if (0 == s1d13xxxfb_pCalcClockBits(ICD_VREG1, S1D13XXX_CLKI, &ulBits))
			return 0;
		if (0 == s1d13xxxfb_pProgramClockChip(ulBits))
			return 0;

		if (0 == s1d13xxxfb_pCalcClockBits(ICD_VREG2, S1D13XXX_CLKI, &ulBits))
			return 0;
		if (0 == s1d13xxxfb_pProgramClockChip(ulBits))
			return 0;
	}

	if (0 != S1D13XXX_CLKI2) {
		if (0 == s1d13xxxfb_pCalcClockBits(ICD_MREG, S1D13XXX_CLKI2, &ulBits))
			return 0;
		if (0 == s1d13xxxfb_pProgramClockChip(ulBits))
			return 0;
	}
	return 1;
}

//---------------------------------------------------------------------------
//  FUNCTION:       s1d13xxxfb_pCalcClockBits()
//
//  DESCRIPTION:
//      This routine calculates the bit pattern required to program the
//      ICD2061A.
//
//  PARAMETERS:
//      Clk             - The input clock to program (i.e. CLKI/CLKI2)
//      ulDesiredFreq   - The desired frequency (in Hz).
//      pulBits         - Pointer to a DWORD to receive the calculated
//                        bit pattern.
//
//  RETURNS:
//      1 if the function was successful in calculatding the bits.
//      0 if there was an error condition detected while trying to
//      calculate the bits.
//
//  MODIFIES:
//---------------------------------------------------------------------------
static int s1d13xxxfb_pCalcClockBits( ICD_REG Reg, u32 ulDesiredFreq, u32 * pulBits )
{
	u32 ulFreq = ulDesiredFreq;      // 
	int M = 0;                          // Post divisor value (0 < D < 15)
	int I;                              // VCO index value. (see ulIndexFreq)
	int P = 0;                          // P and Q counter values
	int Q = 0;
	int fVCO;                           //
	int p, q;                           // Temporary P and Q counter values
	u32 f;                           // Temporary VCO frequency
	u32 diff = 0x7FFFFFFF;           // Seed the difference value
	u32 ulBits;                      // Holder to build the bit pattern into.

	// The VCO freqency (fVCO) must be in the range 50 - 120 MHz.

	// If the requested frequency is greater than 120 MHz we must fail.
	if (ulFreq > 120000000)
		return 0;

	// If the requested frequency is less than 50 MHz then
	// calculate the post VCO-divisor value (M).
	while (ulFreq < (u32)(50000000/(1 << M)))
		M++;
	if (M > 0x07)    // 111b
		return 0;

	ulFreq *= (1 << M);

	// Determine the VCO index field (I).
	// The index field for both VCLK (CLKI) and MCLK (CLKI2) is identical
	// except for 1110b (turn off VCLK) and 1111b (Mux MCLK to VCLK)
	for (I = 0; I < 0x0F; I++)
		if ((ulFreq >= ulIndexFreq[I][0]) && (ulFreq <= ulIndexFreq[I][1]))
			break;

	// DSM note: This formula must exactly match that used in the HAL.  I have checked with Bob
	// to determine that p/q tests should be "<=" instead of "<" (according to the Cypress spec).
	// The Cypress spec states that "3 <= Q <= 129", but in addition, Q is further constrained by 
	// "200KHz <= fREF/Q <= 1MHz".  Since we agreed to use the more stringent constraint, then
	// our Q loop should go from 15 to 71 (inclusive).

	// Calculate the best P and Q counter values.
	for (p = 4; p <= 130; p++)
		for (q = 15; q <= 71; q++) {
			f = (2 * fREF * p)/q;
			if ( (u32) abs((int) f - (int) ulFreq) < diff) {
				diff = abs((int) f - (int) ulFreq);
				P = p; Q = q;
				fVCO = f;
			}
		}

	ulBits = ((Reg & 0x07) << 21) | ((I & 0x0F) << 17) | (((P - 3) & 0x7F) << 10) | ((M & 0x07) << 7) | ((Q - 2) & 0x7F);
	*pulBits = ulBits;

	return 1;
}

//---------------------------------------------------------------------------
//  FUNCTION:   s1d13xxxfb_pProgramClockChip()
//
//  DESCRIPTION:
//      This routine transfers the data bits in ulBits to the ICD2061A.
//
//  PARAMETERS:
//      ulBits  - The bit pattern to be passed to the programmable clock.
//
//  RETURNS:
//
//  MODIFIES:   REG[REG0064_GPIO_STATUS_CONTROL]
//---------------------------------------------------------------------------

static int s1d13xxxfb_pProgramClockChip( u32 ulBits )
{
	u32 ulGpioReg;
	int ii;

	ulGpioReg =s1d13xxxfb_read_reg(REG0064_GPIO_STATUS_CONTROL);

	//  Ensure GPIO1 and GPIO2 are set as outputs, and GPIO3 is set to input.
	ulGpioReg &= ~0x06080006;           // Disable as inputs & set to low
	ulGpioReg |=  0x08060000;           // Enable as outputs
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg);

	// Set clocks and data to high
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg | 0x00000006);

	// Perform the unlock sequence.
	for (ii = 0; ii < 5; ii++)
		WriteClockEdge(1, 1);
	WriteClockEdge(0, 0);

	WriteClockEdge(0, 0);   // Start Bit

	for (ii = 0; ii < 24; ii++) {
	        if (ulBits & 1)
			WriteClockEdge(0, 1);
		else
			WriteClockEdge(1, 0);
		ulBits >>= 1;
	}

	WriteClockEdge(1, 1);  // Stop Bit

	// RELE - we may need to delay here for the ICD2061A to lock down the
	//        changes before we do whatever the next write does.
	// DSM note: My super-duper fast assembly language test I did for the
	// 1356 (or is that the 13506) indicates no delay is required.
    
	// Restore the GPIO Status and Control register (DSM note) but
	// retain GPIO1/2 as outputs and GPIO3 as an input.  Also,
	// GPIO1/2 are driven low to ensure that REG0 is selected.
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg);

	// DSM note: On the SDU13A3B0B eval board, the ICD2061A's INTCLK
	// pin is incorrectly grounded.  This will cause a ZERO frequency
	// on CLKI if GPIO1/2 are set to 0/1.  To ensure this doesn't
	// happen, we set GPIO1/2 to outputs and drive them to 0/0.

	// RELE - we could add a check on the clock status for failure here.
	//      - As this error signal only informs us when we are definately
	//        bad and not when we are for sure good AND because it can be
	//        disconnected on the B0B eval board it is not tested here.
	return 1;
}

//---------------------------------------------------------------------------
//  FUNCTION:       WriteClockEdge()
//
//  DESCRIPTION:
//
//  PARAMETERS:
//
//  RETURNS:
//
//  MODIFIES:   REG[REG64_GPIO_STATUS_CONTROL]
//---------------------------------------------------------------------------
static void WriteClockEdge( int FallingEdgeBit, int RisingEdgeBit )
{
	u32 ulGpioReg =s1d13xxxfb_read_reg(REG0064_GPIO_STATUS_CONTROL) & ~0x06;

	u32 ulFallingEdgeData = (FallingEdgeBit) ? 0x04 : 0;
	u32 ulRisingEdgeData  = (RisingEdgeBit)  ? 0x04 : 0;

	// Write falling edge first
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg | ulFallingEdgeData | 0x02);  // CLK=1
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg | ulFallingEdgeData | 0x00);  // CLK=0

	// Write rising edge next
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg | ulRisingEdgeData | 0x00);   // CLK=0
	s1d13xxxfb_write_reg(REG0064_GPIO_STATUS_CONTROL, ulGpioReg | ulRisingEdgeData | 0x02);   // CLK=1
}

#endif  //CONFIG_FBCON_EPSON_CLOCK_CHIP
