//-----------------------------------------------------------------------------
//  
// linux/drivers/video/epson/s1d13xxxfb_usb.c -- USB 
// functions for Epson S1D13xxx controller frame buffer drivers.
// 
// Copyright(c) Seiko Epson Corporation 2000-2005.
// All rights reserved.
//
// This file is subject to the terms and conditions of the GNU General Public
// License. See the file COPYING in the main directory of this archive for
// more details.
//
//---------------------------------------------------------------------------- 
//---------------------------------------------------------------------------- 
//
// WARNING: This file uses a chip specific include file. The include file 
// contains all hardware information needed to generate proper timings and 
// access registers
//
// This is an USB device driver for reference only.  This driver implements
// Bulk transfers and is using a polling mechanism.  Developers will need
// to implement ISO transfer functions (if necessary) and interrupt handlers.
//
// This driver supports one of the following Epson CRT/TV/LCD controllers:
//
// S1D13A05
//
//-----------------------------------------------------------------------------
#include <linux/version.h>
#include <linux/kernel.h>

#include "s1d13xxxfb.h"
#include "s1d13xxxfb_usb.h"

#ifdef CONFIG_FBCON_EPSON_PCI
#include <linux/timer.h>
#include <linux/jiffies.h>
#endif

#include <linux/smp_lock.h>

#ifdef CONFIG_FBCON_EPSON_USB

//-----------------------------------------------------------------------------
// Local definitions
//----------------------------------------------------------------------------- 

#define EP_MODE_NORMAL      0   //
#define EP_MODE_LOOPBACK    1   // Copies received data to the tranmit EP.

#define SIZEOF_MAIL         8
#define SIZEOF_EP1          SIZEOF_MAIL
#define SIZEOF_FIFO         64

#define USBPOLLRATE         1000

//-----------------------------------------------------------------------------
// Local globals
//----------------------------------------------------------------------------- 
#ifdef CONFIG_FBCON_EPSON_PCI
static struct timer_list usb_irq_poll_timer;
#endif

u16  gfTransferMode = EP_MODE_LOOPBACK; // Loopback or Normal.

u8   gusFifoBuffer[SIZEOF_FIFO];

//-----------------------------------------------------------------------------
// Private Function Prototypes 
//-----------------------------------------------------------------------------
void s1d13xxxfb_WriteUsbExtReg( u16 Index, u8 Data );
void s1d13xxxfb_USBRegInit( void );
void s1d13xxxfb_USBIrqLoop( unsigned long dummy );
void s1d13xxxfb_IrqHandler( void );

void BULK_IrqHandler( void );
void BULK_SuspendHandler( void );
void BULK_SOFHandler( void );
void BULK_EP4Handler( void );
void BULK_EP3Handler( void );
void BULK_EP1Handler( void );

void BULK_LoopMail( void );
void BULK_LoopFifo( void );

void BULK_ReadFifo( void );
void BULK_ReadMail( void );

//---------------------------------------------------------------------------
//  FUNCTION:   s1d13xxxfb_WriteUsbExtReg()
//              WriteUsbExtReg16()
//
//  DESCRIPTION:
//      This function accesses the USB registers located at offset 4038
//      and 403A.
//
//  PARAMETERS:
//      Index   - Offset to the desired extended register.
//      Data    - Informatio to write into the register.
//---------------------------------------------------------------------------
void s1d13xxxfb_WriteUsbExtReg( u16 Index, u8 Data )
{
	s1d13xxxfb_write_reg8(REG4038_USB_EXT_REGISTER_INDEX, Index);
	s1d13xxxfb_write_reg8(REG403A_USB_EXT_REGISTER_DATA, Data);
}   


//---------------------------------------------------------------------------
//  FUNCTION:       s1d13xxxfb_USBInit()
//
//  DESCRIPTION:
//  General entry point for callers to intialize the USB client.
//
//  PARAMETERS:     None.
//
//  RETURNS:        None.
//---------------------------------------------------------------------------
void s1d13xxxfb_USBInit( void )
{
	s1d13xxxfb_USBRegInit();

#ifdef CONFIG_FBCON_EPSON_PCI
	// Create a timer to poll the USB interrupts
	init_timer(&usb_irq_poll_timer);
	usb_irq_poll_timer.function = s1d13xxxfb_USBIrqLoop;
	usb_irq_poll_timer.expires = jiffies+HZ/USBPOLLRATE;
	add_timer(&usb_irq_poll_timer);
#endif
}

//---------------------------------------------------------------------------
//  FUNCTION:       s1d13xxxfb_USBRegInit()
//
//  DESCRIPTION:
//      This routine performs a general purpose setup of the USB registers.
//      There are two intances where this routine should be called.
//
//      First, this routine can be called to init the registers on startup.
//
//      Second, when the USB is plugged into a host controller the host sends
//      a RESET request. After each RESET the contents of the USB registers
//      will be cleared, call this routine to restore the register contents.
//
//  PARAMETERS:     None.
//
//  RETURNS:        None.
//---------------------------------------------------------------------------
void s1d13xxxfb_USBRegInit( void )
{
	u32 regdata;

	// Enable the USB comparator.
	regdata = s1d13xxxfb_read_reg16(REG4040_USB_USBFC_INPUT_CONTROL);
	s1d13xxxfb_write_reg16(REG4040_USB_USBFC_INPUT_CONTROL, 0x0040 | regdata);
	//s1d13xxxfb_write_reg16(REG4040_USB_USBFC_INPUT_CONTROL, 0x0040);

	// USBPUP must be set to enable USB interface and registers.
	s1d13xxxfb_write_reg16(REG4044_USB_PIN_IO_STATUS_DATA, 0x0001);

	// Enable the clocks and USB GPIO pins.
	s1d13xxxfb_write_reg16(REG4000_USB_CONTROL, 0x0084);

	s1d13xxxfb_write_reg16(REG4024_USB_EP3_RX_FIFO_STATUS, 0x001C);  // Clear EP3 status.
	s1d13xxxfb_write_reg16(REG402C_USB_EP4_TX_FIFO_STATUS, 0x001C);  // Clear EP4 status.
	s1d13xxxfb_write_reg16(REG4032_USB_USB_STATUS, 0x007E);          // Clear EP2 valid bit.
	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x00FF);        // Clear ALL USB interrupts.
	s1d13xxxfb_write_reg16(REG4010_USB_EP1_INDEX, 0x000);            // Set EP1 and EP2 indices to zero. 
	s1d13xxxfb_write_reg16(REG4018_USB_EP2_INDEX, 0x0000);

	// Specify our vendor and product IDs.
	s1d13xxxfb_WriteUsbExtReg(0x00, VENDORID_MSB);
	s1d13xxxfb_WriteUsbExtReg(0x01, VENDORID_LSB);
	s1d13xxxfb_WriteUsbExtReg(0x02, PRODUCTID_MSB);
	s1d13xxxfb_WriteUsbExtReg(0x03, PRODUCTID_LSB);

	// Enable EP4 (FIFO) valid transfer mode.
	s1d13xxxfb_WriteUsbExtReg(0x0C, 0x01);

	// Enable interrupts for EP1 and EP3.
	s1d13xxxfb_write_reg16(REG4002_USB_INT_ENABLE_0, 0x000A);
	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x000A);

	// Enable RESET on +ve interrupts and internal INTs on -ve interrupt.
	s1d13xxxfb_write_reg16(REG4046_USB_POS_INT_ENABLE, 0x0002);
	s1d13xxxfb_write_reg16(REG4048_USB_NEG_INT_ENABLE, 0x0001);
	// Clear the external interrupt status registers.
	s1d13xxxfb_write_reg16(REG404A_USB_POS_INT_STATUS, 0x007F);
	s1d13xxxfb_write_reg16(REG404C_USB_NEG_INT_STATUS, 0x007F);

	// Set the wait state register to '1'.
	// This has to be done when using the default BCLK of 50 MHz.
	s1d13xxxfb_write_reg16(REG4054_USB_WAIT_STATE, 0x0001);
    
	// Enable the USB port for use (for non-BULK transfer)
	//regdata = s1d13xxxfb_read_reg16(REG4000_USB_CONTROL);
	//s1d13xxxfb_write_reg16(REG4000_USB_CONTROL, 0x00A4 | regdata);

	// Set BULK transfer mode.
	regdata = s1d13xxxfb_read_reg16(REG4040_USB_USBFC_INPUT_CONTROL);
	s1d13xxxfb_write_reg16(REG4040_USB_USBFC_INPUT_CONTROL, 0x0008 | regdata);
	// Clear the external interrupt status registers.
	// RELE - this step MUST be done after every write to the USBFC Input
	//        Control register.
	s1d13xxxfb_write_reg16(REG404A_USB_POS_INT_STATUS, 0x007F);
	s1d13xxxfb_write_reg16(REG404C_USB_NEG_INT_STATUS, 0x007F);

	// Enable the USB port for use.
	regdata = s1d13xxxfb_read_reg16(REG4000_USB_CONTROL);
	s1d13xxxfb_write_reg16(REG4000_USB_CONTROL, 0x0020 | regdata);
}

#ifdef CONFIG_FBCON_EPSON_PCI
//---------------------------------------------------------------------------
//  FUNCTION:   s1d13xxxfb_USBIrqLoop()
//
//  DESCRIPTION:
//      The evaluation PCI card does not support interrupts.
//      This routine polls the interrupt registers. This is just a reference
//      routine and developers will need to write their own interrupt handlers.
//
//  PARAMETERS:     dummy - dummy argument needed by timer_list.
//
//  RETURNS:        None.
//---------------------------------------------------------------------------
void s1d13xxxfb_USBIrqLoop( unsigned long dummy )
{
	// Check if interrupt has occurred
	if (s1d13xxxfb_read_reg16(REG404E_USB_POS_INT_MASK_STATUS) ||
	    s1d13xxxfb_read_reg16(REG4050_USB_NEG_INT_MASK_STATUS))
		s1d13xxxfb_IrqHandler();

	// Reset the timer
	usb_irq_poll_timer.expires = jiffies+HZ/USBPOLLRATE;
	add_timer(&usb_irq_poll_timer);
}
#endif

//---------------------------------------------------------------------------
//  FUNCTION:   s1d13xxxfb_IrqHandler()
//
//  DESCRIPTION:
//      This is the master interrupt handler. It is called either in response
//      to a real physical interrupt (someday) or when the polling routine
//      detects an interrupt state.
//
//      On receiving interrupt notification this routine determines the
//      type of interrupt and calls the appropriate sub-handler.
//---------------------------------------------------------------------------
void s1d13xxxfb_IrqHandler( void )
{
	// If this is a -ve interrupt then call either BULK_IrqHandler() or ISO_IrqHandler(). 
	if (s1d13xxxfb_read_reg16(REG4050_USB_NEG_INT_MASK_STATUS) & 0x0001) {
		BULK_IrqHandler();
		s1d13xxxfb_write_reg16(REG404C_USB_NEG_INT_STATUS, 0x0001);
		return;
	}

	// If this is +ve interrupt, then we are here due to a USB RESET
	// Re-initialize the USB registers and clear the +ve interrupt register.
	if (s1d13xxxfb_read_reg16(REG404E_USB_POS_INT_MASK_STATUS) & 0x0002) {
		s1d13xxxfb_USBRegInit();
		s1d13xxxfb_write_reg16(REG404A_USB_POS_INT_STATUS, 0x0002);
		return;
	}
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_IrqHandler()
//
//  DESCRIPTION:
//      After the master interupt handler receives an interrrupt either
//      this routine or ISO_IrqHandler() will be called to dispatch control
//      to the appropriate handler routine.
//---------------------------------------------------------------------------
void BULK_IrqHandler( void )
{
	u16 Irq1Status = s1d13xxxfb_read_reg16(REG4002_USB_INT_ENABLE_0) &
	                 s1d13xxxfb_read_reg16(REG4004_USB_INT_STATUS_0);

	if (Irq1Status & 0x80)          // Is this a Suspend interrupt?
		BULK_SuspendHandler();

	if (Irq1Status & 0x40)          // Is this a SOF interrupt?
		BULK_SOFHandler();

	if (Irq1Status & 0x10)          // Is this an EP4 (FIFO TX) interrupt?
		BULK_EP4Handler();

	if (Irq1Status & 0x08)          // Is this an EP3 (FIFO RX) interrupt?
		BULK_EP3Handler();

	if (Irq1Status & 0x02)          // Is this an EP1 (mailbox RX) interrupt?
		BULK_EP1Handler();

	if (Irq1Status & 0x01) {        // Is this an almost full/empty interrupt?
		// Stub
	}
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_SuspendHandler()
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------
void BULK_SuspendHandler( void )
{
	// Stub
}

//---------------------------------------------------------------------------
//  FUNCTION:   BULK_SOFHandler()
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------
void BULK_SOFHandler( void )
{
	// Stub

	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x0040);
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_EP4Handler()
//
//  DESCRIPTION:
//---------------------------------------------------------------------------
void BULK_EP4Handler( void )
{
	switch (gfTransferMode) {
	case EP_MODE_NORMAL:
		break;
	case EP_MODE_LOOPBACK:
		printk("EP4 IRQ during loopback!!");
		break;
	default:
		break;
	}

	// Reset the interrupt status.
	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x0010);
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_EP3Handler()
//
//  DESCRIPTION:
//      EP3 has just interrupted. That means the EP3 FIFO is full and needs
//      to be serviced.
//---------------------------------------------------------------------------
void BULK_EP3Handler( void )
{
	int ii;

	if (s1d13xxxfb_read_reg16(REG4032_USB_USB_STATUS) & 0x0002) {         // Did we ACK?
		switch (gfTransferMode) {
		case EP_MODE_NORMAL:
			BULK_ReadFifo();
			break;
		case EP_MODE_LOOPBACK:
			BULK_LoopFifo();
			break;
		default:
			break;
		}
	} else if (0 == (s1d13xxxfb_read_reg16(REG4032_USB_USB_STATUS) & 0x0004))  // Did we NAK?
		s1d13xxxfb_write_reg16(REG4024_USB_EP3_RX_FIFO_STATUS, 0x0010);      // No - Flush the EP3 FIFO.

	s1d13xxxfb_write_reg16(REG4032_USB_USB_STATUS, 0x06);                    // Reset ACK/NAK bits.

	// This little time waster is from NC2888 Errata 630-0002-0304
	// Check the NAK status for 50us. If we find it set we can exit early.
	// (I hope 10 passes is 50us - or more)
	for (ii = 0; ii < 10; ii++)
		if (s1d13xxxfb_read_reg16(REG4032_USB_USB_STATUS) & 0x0004)
			break;

	s1d13xxxfb_write_reg16(REG4032_USB_USB_STATUS, 0x06);  // Reset ACK/NAK status bits.

	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x08);  // Reset the EP3 interrupt.
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_EP1Handler()
//
//  DESCRIPTION:
//      Called in response the interrupt handler receiving an EP1 (mailbox)
//      interrupt (which means the host has sent data to us using EP1)
//
//      Call the approprite routine, based on our opearating mode, then
//      reset the interrupt status.
//---------------------------------------------------------------------------
void BULK_EP1Handler( void )
{
	switch (gfTransferMode) {
	case EP_MODE_NORMAL:
		BULK_ReadMail();
		break;
	case EP_MODE_LOOPBACK:
		BULK_LoopMail();
		break;
	default:
		break;
	}

	// Reset the interrupt.
	s1d13xxxfb_write_reg16(REG4004_USB_INT_STATUS_0, 0x02);
}   

//---------------------------------------------------------------------------
//  FUNCTION:   BULK_LoopMailBox()
//
//  DESCRIPTION:
//      This routine is called when we are in loopback mode and an EP1
//      interrupt is received. The EP1 (mailbox receive) data register
//      is read SIZEOF_MAILBOX times and the data is written to the EP2
//      (mailbox transmit) data register.
//---------------------------------------------------------------------------
void BULK_LoopMail( void )
{
	int ii;

	// Check EP2 (mailbox) status. If the mailbox was written by the local
	// CPU but has not been read by the host then there's nothing to do.
	if (!(1 & s1d13xxxfb_read_reg16(REG4032_USB_USB_STATUS))) {
		for (ii = 0; ii < SIZEOF_MAIL; ii++)
			s1d13xxxfb_write_reg16(REG401A_USB_EP2_TX_MAILBOX_DATA, s1d13xxxfb_read_reg16(REG4012_USB_EP1_RX_MAILBOX_DATA));
		// Set EP2 (transmit mailbox) valid.
		s1d13xxxfb_write_reg16(REG4032_USB_USB_STATUS, 0x0001);
	}
}   

//---------------------------------------------------------------------------
//  FUNCTION:   BULK_LoopFifo()
//
//  DESCRIPTION:
//      This routine is called when an EP3 (FIFO receive) interrupt is received
//      and we are in loopback mode.
//
//      The EP3 FIFO count register is read to retrieve the number of FIFO
//      entries. Then the EP3 Data register is read the indicated number of
//      times and the data read is placed into the EP4 (FIFO transmit) data
//      register.
//
//      Finally the EP4 (FIFO transmit) valid bit is set.
//---------------------------------------------------------------------------
void BULK_LoopFifo( void )
{
	int ii;
	int nCount;

	// Has the FIFO transmit buffer been read by the host yet?
	if (s1d13xxxfb_read_reg16(REG402C_USB_EP4_TX_FIFO_STATUS) & 0x20)
		printk("FIFO loopback ERROR: TX FIFO not empty!\a\n");
	else {
		nCount = s1d13xxxfb_read_reg16(REG4022_USB_EP3_RX_FIFO_COUNT);

		for (ii = 0; ii < nCount; ii++)
			s1d13xxxfb_write_reg16(REG4028_USB_EP4_TX_FIFO_DATA, s1d13xxxfb_read_reg16(REG4020_USB_EP3_RX_FIFO_DATA));

		s1d13xxxfb_write_reg16(REG402C_USB_EP4_TX_FIFO_STATUS, 0x20);    // Set EP4 (transmit FIFO) valid bit.
	}
}

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_ReadMail()
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------
void BULK_ReadMail( void )
{
	// Stub
}   

//---------------------------------------------------------------------------
//  FUNCTION:       BULK_ReadFifo()
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------
void BULK_ReadFifo( void )
{
	// Stub
}   

#endif  //CONFIG_FBCON_EPSON_USB
